@@ -34,6 +34,7 @@ module Control.Concurrent.STM.TBQueue (
         newTBQueueIO,
         readTBQueue,
         tryReadTBQueue,
+        flushTBQueue,
         peekTBQueue,
         tryPeekTBQueue,
         writeTBQueue,
@@ -101,12 +102,12 @@ writeTBQueue :: TBQueue a -> a -> STM ()
 writeTBQueue (TBQueue rsize _read wsize write) a = do
   w <- readTVar wsize
   if (w /= 0)
-     then do writeTVar wsize (w - 1)
+     then do writeTVar wsize $! w - 1
      else do
           r <- readTVar rsize
           if (r /= 0)
              then do writeTVar rsize 0
-                     writeTVar wsize (r - 1)
+                     writeTVar wsize $! r - 1
              else retry
   listend <- readTVar write
   writeTVar write (a:listend)
@@ -116,7 +117,7 @@ readTBQueue :: TBQueue a -> STM a
 readTBQueue (TBQueue rsize read _wsize write) = do
   xs <- readTVar read
   r <- readTVar rsize
-  writeTVar rsize (r + 1)
+  writeTVar rsize $! r + 1
   case xs of
     (x:xs') -> do
       writeTVar read xs'
@@ -137,6 +138,22 @@ readTBQueue (TBQueue rsize read _wsize write) = do
 tryReadTBQueue :: TBQueue a -> STM (Maybe a)
 tryReadTBQueue c = fmap Just (readTBQueue c) `orElse` return Nothing
 
+-- | Efficiently read the entire contents of a 'TBQueue' into a list. This
+-- function never retries.
+--
+-- @since 2.4.5
+flushTBQueue :: TBQueue a -> STM [a]
+flushTBQueue (TBQueue rsize read wsize write) = do
+  xs <- readTVar read
+  ys <- readTVar write
+  r <- readTVar rsize
+  w <- readTVar wsize
+  writeTVar read []
+  writeTVar write []
+  writeTVar rsize 0
+  writeTVar wsize (r + w)
+  return (xs ++ reverse ys)
+
 -- | Get the next value from the @TBQueue@ without removing it,
 -- retrying if the channel is empty.
 peekTBQueue :: TBQueue a -> STM a
@@ -162,11 +179,11 @@ unGetTBQueue :: TBQueue a -> a -> STM ()
 unGetTBQueue (TBQueue rsize read wsize _write) a = do
   r <- readTVar rsize
   if (r > 0)
-     then do writeTVar rsize (r - 1)
+     then do writeTVar rsize $! r - 1
      else do
           w <- readTVar wsize
           if (w > 0)
-             then writeTVar wsize (w - 1)
+             then writeTVar wsize $! w - 1
              else retry
   xs <- readTVar read
   writeTVar read (a:xs)
